//#define LOG_NDEBUG 0
#define LOG_TAG "SwConverter"
#include <log/log.h>

#include <stdlib.h>
#include <string.h>
#include <libyuv.h>
#include "Debug.h"
#include "SwConverter.h"

static int sw_yuv422_to_yuv422_32Align = 0;
static int sh_yuv422_to_yuv422_32Align = 0;
void yuv422_to_yuv422_32Align(unsigned char* bufsrc, unsigned char* bufdest, int width, int height)
{
    if ((width != sw_yuv422_to_yuv422_32Align) || (height != sh_yuv422_to_yuv422_32Align) ) {
        ALOGV("%s: image size: src:[%ix%i]\n", __FUNCTION__, width, height);
        sw_yuv422_to_yuv422_32Align = width;
        sh_yuv422_to_yuv422_32Align = height;
    }
    int destWidth = ((width + 31) >> 5) << 5;
    int destSize = destWidth * height * 2;

    memset(bufdest, 0, destSize);
    for (int i = 0; i < height; ++ i) {
        memcpy(bufdest + destWidth * 2 * i, bufsrc + width * 2 * i, width * 2);
    }
}

void
yuyv422_to_yuv420(unsigned char* bufsrc, unsigned char* bufdest, int width, int height)
{
    unsigned char* ptrsrcy1, *ptrsrcy2;
    unsigned char* ptrsrcy3, *ptrsrcy4;
    unsigned char* ptrsrccb1, *ptrsrccb2;
    unsigned char* ptrsrccb3, *ptrsrccb4;
    unsigned char* ptrsrccr1, *ptrsrccr2;
    unsigned char* ptrsrccr3, *ptrsrccr4;
    int srcystride, srcccstride;

    ptrsrcy1  = bufsrc ;
    ptrsrcy2  = bufsrc + (width << 1) ;
    ptrsrcy3  = bufsrc + (width << 1) * 2 ;
    ptrsrcy4  = bufsrc + (width << 1) * 3 ;

    ptrsrccb1 = bufsrc + 1;
    ptrsrccb2 = bufsrc + (width << 1) + 1;
    ptrsrccb3 = bufsrc + (width << 1) * 2 + 1;
    ptrsrccb4 = bufsrc + (width << 1) * 3 + 1;

    ptrsrccr1 = bufsrc + 3;
    ptrsrccr2 = bufsrc + (width << 1) + 3;
    ptrsrccr3 = bufsrc + (width << 1) * 2 + 3;
    ptrsrccr4 = bufsrc + (width << 1) * 3 + 3;

    srcystride  = (width << 1) * 3;
    srcccstride = (width << 1) * 3;

    unsigned char* ptrdesty1, *ptrdesty2;
    unsigned char* ptrdesty3, *ptrdesty4;
    unsigned char* ptrdestcb1, *ptrdestcb2;
    unsigned char* ptrdestcr1, *ptrdestcr2;
    int destystride, destccstride;

    ptrdesty1 = bufdest;
    ptrdesty2 = bufdest + width;
    ptrdesty3 = bufdest + width * 2;
    ptrdesty4 = bufdest + width * 3;

    ptrdestcb1 = bufdest + width * height;
    ptrdestcb2 = bufdest + width * height + (width >> 1);

    ptrdestcr1 = bufdest + width * height + ((width * height) >> 2);
    ptrdestcr2 = bufdest + width * height + ((width * height) >> 2) + (width >> 1);

    destystride  = (width) * 3;
    destccstride = (width >> 1);

    int i, j;

    for(j = 0; j < (height / 4); j++) {
        for(i = 0; i < (width / 2); i++) {
            (*ptrdesty1++) = (*ptrsrcy1);
            (*ptrdesty2++) = (*ptrsrcy2);
            (*ptrdesty3++) = (*ptrsrcy3);
            (*ptrdesty4++) = (*ptrsrcy4);

            ptrsrcy1 += 2;
            ptrsrcy2 += 2;
            ptrsrcy3 += 2;
            ptrsrcy4 += 2;

            (*ptrdesty1++) = (*ptrsrcy1);
            (*ptrdesty2++) = (*ptrsrcy2);
            (*ptrdesty3++) = (*ptrsrcy3);
            (*ptrdesty4++) = (*ptrsrcy4);

            ptrsrcy1 += 2;
            ptrsrcy2 += 2;
            ptrsrcy3 += 2;
            ptrsrcy4 += 2;

            (*ptrdestcb1++) = (*ptrsrccb1);
            (*ptrdestcb2++) = (*ptrsrccb3);

            ptrsrccb1 += 4;
            ptrsrccb3 += 4;

            (*ptrdestcr1++) = (*ptrsrccr1);
            (*ptrdestcr2++) = (*ptrsrccr3);

            ptrsrccr1 += 4;
            ptrsrccr3 += 4;

        }


        /* Update src pointers */
        ptrsrcy1  += srcystride;
        ptrsrcy2  += srcystride;
        ptrsrcy3  += srcystride;
        ptrsrcy4  += srcystride;

        ptrsrccb1 += srcccstride;
        ptrsrccb3 += srcccstride;

        ptrsrccr1 += srcccstride;
        ptrsrccr3 += srcccstride;


        /* Update dest pointers */
        ptrdesty1 += destystride;
        ptrdesty2 += destystride;
        ptrdesty3 += destystride;
        ptrdesty4 += destystride;

        ptrdestcb1 += destccstride;
        ptrdestcb2 += destccstride;

        ptrdestcr1 += destccstride;
        ptrdestcr2 += destccstride;

    }
}

void yuyv422_to_yuv420sp_libyuv(bool bNV21, unsigned char* bufsrc, unsigned char* bufdest, int width, int height)
{
    unsigned int YSize = width * height;
    size_t src_size = YSize * 1.5;
    unsigned char* I420 = new unsigned char[(int)src_size];

    unsigned char* pI420Y = I420;
    unsigned char* pI420U = I420 + YSize;
    unsigned char* pI420V = pI420U + (YSize / 4);

    unsigned char* pDstY = bufdest;
    unsigned char* pDstUV = bufdest + YSize;

    int ret = libyuv::ConvertToI420(bufsrc, src_size, pI420Y, width, pI420U, width >> 1, pI420V, width >> 1,
                                    0, 0, width, height, width, height, libyuv::kRotate0, libyuv::FOURCC_YUY2);

    // XCheck: Merlin3 use I420ToNV12, Mac5p use I420ToNV21
    if (bNV21) {
        ret = libyuv::I420ToNV21(pI420Y, width, pI420U, width >> 1, pI420V, width >> 1, pDstY, width, pDstUV, width, width, height); // Mac5p
    } else {
        ret = libyuv::I420ToNV12(pI420Y, width, pI420U, width >> 1, pI420V, width >> 1, pDstY, width, pDstUV, width, width, height); // Merlin3
    }
    if (I420) {
        delete[] I420;
    }
}

void yuyv422_to_yuv420p_libyuv(unsigned char* bufsrc, unsigned char* bufdestY, unsigned char* bufdestCb, unsigned char* bufdestCr, int width, int height)
{
    unsigned int YSize = width * height;
    size_t src_size = YSize * 1.5;

    int ret = libyuv::ConvertToI420(bufsrc, src_size, bufdestY, width, bufdestCb, width >> 1, bufdestCr, width >> 1,
                                    0, 0, width, height, width, height, libyuv::kRotate0, libyuv::FOURCC_YUY2);
}

void yuv420sp_to_yuyv422_libyuv(unsigned char* bufsrc, unsigned char* bufdest, int width, int height)
{
    unsigned int YSize = width * height;
    size_t src_size = YSize * 1.5;
    unsigned char* I420 = new unsigned char[(int)src_size];

    unsigned char* pI420Y = I420;
    unsigned char* pI420U = I420 + YSize;
    unsigned char* pI420V = pI420U + (YSize / 4);

    unsigned char* pDstYUY2 = bufdest;

    int ret = libyuv::ConvertToI420(bufsrc, src_size,
                                    pI420Y, width,
                                    pI420U, width >> 1,
                                    pI420V, width >> 1,
                                    0, 0,//crop x,y
                                    width, height,
                                    width, height,//crop width,height
                                    libyuv::kRotate0,
                                    libyuv::FOURCC_NV12);

    ret = libyuv::I420ToYUY2(
              pI420Y, width,
              pI420U, width >> 1,
              pI420V, width >> 1,
              pDstYUY2, width << 1,
              width, height);
    if (I420) {
        delete[] I420;
    }
}

void yuyv422_to_yuv420sp(unsigned char* bufsrc, unsigned char* bufdest, int width, int height)
{
    unsigned char* ptrsrcy1, *ptrsrcy2;
    unsigned char* ptrsrcy3, *ptrsrcy4;
    unsigned char* ptrsrccb1, *ptrsrccb2;
    unsigned char* ptrsrccb3, *ptrsrccb4;
    unsigned char* ptrsrccr1, *ptrsrccr2;
    unsigned char* ptrsrccr3, *ptrsrccr4;
    int srcystride, srcccstride;

    ptrsrcy1  = bufsrc ;
    ptrsrcy2  = bufsrc + (width << 1) ;
    ptrsrcy3  = bufsrc + (width << 1) * 2 ;
    ptrsrcy4  = bufsrc + (width << 1) * 3 ;

    ptrsrccb1 = bufsrc + 1;
    ptrsrccb2 = bufsrc + (width << 1) + 1;
    ptrsrccb3 = bufsrc + (width << 1) * 2 + 1;
    ptrsrccb4 = bufsrc + (width << 1) * 3 + 1;

    ptrsrccr1 = bufsrc + 3;
    ptrsrccr2 = bufsrc + (width << 1) + 3;
    ptrsrccr3 = bufsrc + (width << 1) * 2 + 3;
    ptrsrccr4 = bufsrc + (width << 1) * 3 + 3;

    srcystride  = (width << 1) * 3;
    srcccstride = (width << 1) * 3;

    unsigned char* ptrdesty1, *ptrdesty2;
    unsigned char* ptrdesty3, *ptrdesty4;
    unsigned char* ptrdestcb1, *ptrdestcb2;
    unsigned char* ptrdestcr1, *ptrdestcr2;
    int destystride, destccstride;

    ptrdesty1 = bufdest;
    ptrdesty2 = bufdest + width;
    ptrdesty3 = bufdest + width * 2;
    ptrdesty4 = bufdest + width * 3;

#if 1 // switch cb and cr
    ptrdestcb1 = bufdest + width * height;
    ptrdestcb2 = bufdest + width * height + width;

    ptrdestcr1 = bufdest + width * height + 1;
    ptrdestcr2 = bufdest + width * height + width + 1;
#else
    ptrdestcr1 = bufdest + width * height;
    ptrdestcr2 = bufdest + width * height + width;

    ptrdestcb1 = bufdest + width * height + 1;
    ptrdestcb2 = bufdest + width * height + width + 1;
#endif

    destystride  = (width) * 3;
    destccstride = width;

    int i, j;

    for(j = 0; j < (height / 4); j++) {
        for(i = 0; i < (width / 2); i++) {
            (*ptrdesty1++) = (*ptrsrcy1);
            (*ptrdesty2++) = (*ptrsrcy2);
            (*ptrdesty3++) = (*ptrsrcy3);
            (*ptrdesty4++) = (*ptrsrcy4);

            ptrsrcy1 += 2;
            ptrsrcy2 += 2;
            ptrsrcy3 += 2;
            ptrsrcy4 += 2;

            (*ptrdesty1++) = (*ptrsrcy1);
            (*ptrdesty2++) = (*ptrsrcy2);
            (*ptrdesty3++) = (*ptrsrcy3);
            (*ptrdesty4++) = (*ptrsrcy4);

            ptrsrcy1 += 2;
            ptrsrcy2 += 2;
            ptrsrcy3 += 2;
            ptrsrcy4 += 2;

            (*ptrdestcb1) = (*ptrsrccb1);
            (*ptrdestcb2) = (*ptrsrccb3);
            ptrdestcb1 += 2;
            ptrdestcb2 += 2;

            ptrsrccb1 += 4;
            ptrsrccb3 += 4;

            (*ptrdestcr1) = (*ptrsrccr1);
            (*ptrdestcr2) = (*ptrsrccr3);
            ptrdestcr1 += 2;
            ptrdestcr2 += 2;

            ptrsrccr1 += 4;
            ptrsrccr3 += 4;

        }


        /* Update src pointers */
        ptrsrcy1  += srcystride;
        ptrsrcy2  += srcystride;
        ptrsrcy3  += srcystride;
        ptrsrcy4  += srcystride;

        ptrsrccb1 += srcccstride;
        ptrsrccb3 += srcccstride;

        ptrsrccr1 += srcccstride;
        ptrsrccr3 += srcccstride;


        /* Update dest pointers */
        ptrdesty1 += destystride;
        ptrdesty2 += destystride;
        ptrdesty3 += destystride;
        ptrdesty4 += destystride;

        ptrdestcb1 += destccstride;
        ptrdestcb2 += destccstride;

        ptrdestcr1 += destccstride;
        ptrdestcr2 += destccstride;

    }
}

void yuyv422_to_yuv420sp_libyuv_scale(bool bNV21, unsigned char* pSrc, int srcWidth, int srcHeight, unsigned char* pDst, int dstWidth, int dstHeight)
{
    unsigned int YSize = srcWidth * srcHeight;
    size_t src_size = YSize * 1.5;
    unsigned char* I420 = new unsigned char[(int)src_size];

    unsigned char* pI420Y = I420;
    unsigned char* pI420U = I420 + YSize;
    unsigned char* pI420V = pI420U + (YSize / 4);

    int ret = libyuv::ConvertToI420(pSrc, src_size,
                                    pI420Y, srcWidth,
                                    pI420U, srcWidth >> 1,
                                    pI420V, srcWidth >> 1,
                                    0, 0,//crop x,y
                                    srcWidth, srcHeight,
                                    srcWidth, srcHeight,//crop width,height
                                    libyuv::kRotate0,
                                    libyuv::FOURCC_YUY2);

    unsigned int YSize_scaled = dstWidth * dstHeight;
    size_t src_size_scaled = YSize_scaled * 1.5;
    unsigned char* I420_scaled = new unsigned char[(int)src_size_scaled];

    unsigned char* pI420Y_scaled = I420_scaled;
    unsigned char* pI420U_scaled = I420_scaled + YSize_scaled;
    unsigned char* pI420V_scaled = pI420U_scaled + (YSize_scaled / 4);

    ret = libyuv::I420Scale(
              pI420Y, srcWidth,
              pI420U, srcWidth >> 1,
              pI420V, srcWidth >> 1,
              srcWidth, srcHeight,
              pI420Y_scaled, dstWidth,
              pI420U_scaled, dstWidth >> 1,
              pI420V_scaled, dstWidth >> 1,
              dstWidth, dstHeight,
              libyuv::kFilterNone);

    unsigned char* pDstY = pDst;
    unsigned char* pDstUV = pDst + YSize_scaled;

    if (bNV21) {
        ret = libyuv::I420ToNV21(pI420Y_scaled, dstWidth, pI420U_scaled, dstWidth >> 1, pI420V_scaled, dstWidth >> 1, pDstY, dstWidth, pDstUV, dstWidth, dstWidth, dstHeight);
    } else {
        ret = libyuv::I420ToNV12(pI420Y_scaled, dstWidth, pI420U_scaled, dstWidth >> 1, pI420V_scaled, dstWidth >> 1, pDstY, dstWidth, pDstUV, dstWidth, dstWidth, dstHeight);
    }

    if (I420) {
        delete[] I420;
    }
    if (I420_scaled) {
        delete[] I420_scaled;
    }
}

static int sw = 0;
static int sh = 0;
static int dw = 0;
static int dh = 0;
void yuyv422_to_yuv420sp_scale(unsigned char* pSrc, int srcWidth, int srcHeight, unsigned char* pDst, int dstWidth, int dstHeight)
{
    if ((srcWidth < dstWidth) || (srcHeight < dstHeight)
        || (srcWidth != (((srcWidth + 3) >> 2) << 2))
        || (srcHeight != (((srcHeight + 3) >> 2) << 2))
        || (dstWidth != (((dstWidth + 3) >> 2) << 2))
        || (dstHeight != (((dstHeight + 3) >> 2) << 2))) {
        ALOGE("%s: image size error!!! src:[%ix%i], dst:[%ix%i]\n", __FUNCTION__, srcWidth, srcHeight, dstWidth, dstHeight);
    }
    if ((srcWidth != sw) || (srcHeight != sh) || (dstWidth != dw) || (dstHeight != dh)) {
        ALOGV("%s: image size: src:[%ix%i], dst:[%ix%i]\n", __FUNCTION__, srcWidth, srcHeight, dstWidth, dstHeight);
        sw = srcWidth;
        sh = srcHeight;
        dw = dstWidth;
        dh = dstHeight;
    }

    unsigned char* ptrsrcy1, *ptrsrcy2;
    unsigned char* ptrsrcy3, *ptrsrcy4;
    unsigned char* ptrsrccb1, *ptrsrccb2;
    unsigned char* ptrsrccb3, *ptrsrccb4;
    unsigned char* ptrsrccr1, *ptrsrccr2;
    unsigned char* ptrsrccr3, *ptrsrccr4;
    int srcystride, srcccstride;

    ptrsrcy1  = pSrc ;
    ptrsrcy2  = pSrc + (srcWidth << 1) ;
    ptrsrcy3  = pSrc + (srcWidth << 1) * 2 ;
    ptrsrcy4  = pSrc + (srcWidth << 1) * 3 ;

    ptrsrccb1 = pSrc + 1;
    ptrsrccb2 = pSrc + (srcWidth << 1) + 1;
    ptrsrccb3 = pSrc + (srcWidth << 1) * 2 + 1;
    ptrsrccb4 = pSrc + (srcWidth << 1) * 3 + 1;

    ptrsrccr1 = pSrc + 3;
    ptrsrccr2 = pSrc + (srcWidth << 1) + 3;
    ptrsrccr3 = pSrc + (srcWidth << 1) * 2 + 3;
    ptrsrccr4 = pSrc + (srcWidth << 1) * 3 + 3;

    srcystride  = (srcWidth << 1) * 3;
    srcccstride = (srcWidth << 1) * 3;

    unsigned char* ptrdesty1, *ptrdesty2;
    unsigned char* ptrdesty3, *ptrdesty4;
    unsigned char* ptrdestcb1, *ptrdestcb2;
    unsigned char* ptrdestcr1, *ptrdestcr2;
    int destystride, destccstride;

    ptrdesty1 = pDst;
    ptrdesty2 = pDst + dstWidth;
    ptrdesty3 = pDst + dstWidth * 2;
    ptrdesty4 = pDst + dstWidth * 3;

#if 1 // switch cb and cr
    ptrdestcb1 = pDst + dstWidth * dstHeight;
    ptrdestcb2 = pDst + dstWidth * dstHeight + dstWidth;

    ptrdestcr1 = pDst + dstWidth * dstHeight + 1;
    ptrdestcr2 = pDst + dstWidth * dstHeight + dstWidth + 1;
#else
    ptrdestcr1 = pDst + dstWidth * dstHeight;
    ptrdestcr2 = pDst + dstWidth * dstHeight + dstWidth;

    ptrdestcb1 = pDst + dstWidth * dstHeight + 1;
    ptrdestcb2 = pDst + dstWidth * dstHeight + dstWidth + 1;
#endif

    destystride  = (dstWidth) * 3;
    destccstride = dstWidth;

    int srcWIndex = 0;
    int dstWIndex = 0;
    int calDstWIndex =  0;
    bool skipDstWIndex = false;
    int srcHIndex = 0;
    int dstHIndex = 0;
    int calDstHIndex =  0;
    bool skipDstHIndex = false;
    int i, j;
    for(j = 0; j < (srcHeight / 4); j++) {
        if (!skipDstHIndex) {
            srcWIndex = 0;
            dstWIndex = 0;
            skipDstWIndex = false;
            for(i = 0; i < (srcWidth / 2); i++) {
                if (!skipDstWIndex) {
                    (*ptrdesty1++) = (*ptrsrcy1);
                    (*ptrdesty2++) = (*ptrsrcy2);
                    (*ptrdesty3++) = (*ptrsrcy3);
                    (*ptrdesty4++) = (*ptrsrcy4);
                }

                ptrsrcy1 += 2;
                ptrsrcy2 += 2;
                ptrsrcy3 += 2;
                ptrsrcy4 += 2;

                if (!skipDstWIndex) {
                    (*ptrdesty1++) = (*ptrsrcy1);
                    (*ptrdesty2++) = (*ptrsrcy2);
                    (*ptrdesty3++) = (*ptrsrcy3);
                    (*ptrdesty4++) = (*ptrsrcy4);
                }

                ptrsrcy1 += 2;
                ptrsrcy2 += 2;
                ptrsrcy3 += 2;
                ptrsrcy4 += 2;

                if (!skipDstWIndex) {
                    (*ptrdestcb1) = (*ptrsrccb1);
                    (*ptrdestcb2) = (*ptrsrccb3);
                    ptrdestcb1 += 2;
                    ptrdestcb2 += 2;
                }

                ptrsrccb1 += 4;
                ptrsrccb3 += 4;

                if (!skipDstWIndex) {
                    (*ptrdestcr1) = (*ptrsrccr1);
                    (*ptrdestcr2) = (*ptrsrccr3);
                    ptrdestcr1 += 2;
                    ptrdestcr2 += 2;
                }

                ptrsrccr1 += 4;
                ptrsrccr3 += 4;

                srcWIndex += 2;
                if (!skipDstWIndex) {
                    dstWIndex += 2;
                }
                calDstWIndex = (int)((srcWIndex) * dstWidth / srcWidth);
                if (calDstWIndex < dstWIndex) { // skip next dst index
                    skipDstWIndex = true;
                } else {
                    skipDstWIndex = false;
                }
            }
        }

        /* Update src pointers */
        ptrsrcy1  += srcystride;
        ptrsrcy2  += srcystride;
        ptrsrcy3  += srcystride;
        ptrsrcy4  += srcystride;

        ptrsrccb1 += srcccstride;
        ptrsrccb3 += srcccstride;

        ptrsrccr1 += srcccstride;
        ptrsrccr3 += srcccstride;


        /* Update dest pointers */
        if (!skipDstHIndex) {
            ptrdesty1 += destystride;
            ptrdesty2 += destystride;
            ptrdesty3 += destystride;
            ptrdesty4 += destystride;

            ptrdestcb1 += destccstride;
            ptrdestcb2 += destccstride;

            ptrdestcr1 += destccstride;
            ptrdestcr2 += destccstride;
        }

        srcHIndex += 4;
        if (!skipDstHIndex) {
            dstHIndex += 4;
        }
        calDstHIndex = (int)((srcHIndex) * dstHeight / srcHeight);
        if (calDstHIndex < dstHIndex) { // skip next dst index
            skipDstHIndex = true;
        } else {
            skipDstHIndex = false;
        }
    }
}

static int sw2 = 0;
static int sh2 = 0;
static int dw2 = 0;
static int dh2 = 0;

void yuyv422_scale_libyuv(unsigned char* pSrc, int srcWidth, int srcHeight, unsigned char* pDst, int dstWidth, int dstHeight)
{
    LOG_FUNCTION_NAME();
    unsigned int YSize = srcWidth * srcHeight;
    size_t src_size = YSize * 1.5;
    unsigned char* I420 = new unsigned char[(int)src_size];

    unsigned char* pI420Y = I420;
    unsigned char* pI420U = I420 + YSize;
    unsigned char* pI420V = pI420U + (YSize / 4);

    int ret = libyuv::ConvertToI420(pSrc, src_size,
                                    pI420Y, srcWidth,
                                    pI420U, srcWidth >> 1,
                                    pI420V, srcWidth >> 1,
                                    0, 0,//crop x,y
                                    srcWidth, srcHeight,
                                    srcWidth, srcHeight,//crop width,height
                                    libyuv::kRotate0,
                                    libyuv::FOURCC_YUY2);

    unsigned int YSize_scaled = dstWidth * dstHeight;
    size_t src_size_scaled = YSize_scaled * 1.5;
    unsigned char* I420_scaled = new unsigned char[(int)src_size_scaled];

    unsigned char* pI420Y_scaled = I420_scaled;
    unsigned char* pI420U_scaled = I420_scaled + YSize_scaled;
    unsigned char* pI420V_scaled = pI420U_scaled + (YSize_scaled / 4);

    ret = libyuv::I420Scale(
              pI420Y, srcWidth,
              pI420U, srcWidth >> 1,
              pI420V, srcWidth >> 1,
              srcWidth, srcHeight,
              pI420Y_scaled, dstWidth,
              pI420U_scaled, dstWidth >> 1,
              pI420V_scaled, dstWidth >> 1,
              dstWidth, dstHeight,
              libyuv::kFilterNone);

    ret = libyuv::I420ToYUY2(
              pI420Y_scaled, dstWidth,
              pI420U_scaled, dstWidth >> 1,
              pI420V_scaled, dstWidth >> 1,
              pDst, dstWidth << 1,
              dstWidth, dstHeight);

    if (I420) {
        delete[] I420;
    }
    if (I420_scaled) {
        delete[] I420_scaled;
    }
}

void yuyv422_scale(unsigned char* pSrc, int srcWidth, int srcHeight, unsigned char* pDst, int dstWidth, int dstHeight)
{
    if ((srcWidth <= 0) || (srcHeight <= 0) || (dstWidth <= 0) || (dstHeight <= 0) || (pSrc == NULL) || (pDst == NULL)) {
        ALOGE("%s[%d]: image size error!!!Return. src:[%ix%i], dst:[%ix%i], pSrc:%p, pDst:%p\n", __FUNCTION__, __LINE__, srcWidth, srcHeight, dstWidth, dstHeight, pSrc, pDst);
        return;
    }
    if ((srcWidth < dstWidth) || (srcHeight < dstHeight)
        || (srcWidth != (((srcWidth + 3) >> 2) << 2))
        || (srcHeight != (((srcHeight + 3) >> 2) << 2))
        || (dstWidth != (((dstWidth + 3) >> 2) << 2))
        || (dstHeight != (((dstHeight + 3) >> 2) << 2))) {
        ALOGE("%s[%d]: image size error!!! src:[%ix%i], dst:[%ix%i]\n", __FUNCTION__, __LINE__, srcWidth, srcHeight, dstWidth, dstHeight);
    }
    if ((srcWidth != sw2) || (srcHeight != sh2) || (dstWidth != dw2) || (dstHeight != dh2)) {
        ALOGV("%s[%d]: image size: src:[%ix%i], dst:[%ix%i]\n", __FUNCTION__, __LINE__, srcWidth, srcHeight, dstWidth, dstHeight);
        sw2 = srcWidth;
        sh2 = srcHeight;
        dw2 = dstWidth;
        dh2 = dstHeight;
    }

    unsigned char* ptrsrcy1;
    unsigned char* ptrsrccb1;
    unsigned char* ptrsrccr1;
    unsigned char* ptrdesty1;
    unsigned char* ptrdestcb1;
    unsigned char* ptrdestcr1;

    ptrsrcy1  = pSrc ;
    ptrsrccb1 = pSrc + 1;
    ptrsrccr1 = pSrc + 3;
    ptrdesty1 = pDst;
    ptrdestcb1 = pDst + 1;
    ptrdestcr1 = pDst + 3;

    int srcystride  = srcWidth << 1;

    int srcWIndex = 0;
    int dstWIndex = 0;
    int calDstWIndex =  0;
    bool skipDstWIndex = false;

    int srcHIndex = 0;
    int dstHIndex = 0;
    int calDstHIndex =  0;
    bool skipDstHIndex = false;

    int i, j;
    for(j = 0; j < srcHeight; j++) {
        if (!skipDstHIndex) {
            srcWIndex = 0;
            dstWIndex = 0;
            skipDstWIndex = false;
            for(i = 0; i < (srcWidth / 2); i++) {
                if (!skipDstWIndex) {
                    (*ptrdesty1) = (*ptrsrcy1);
                    ptrdesty1 += 2;
                }
                ptrsrcy1 += 2;

                if (!skipDstWIndex) {
                    (*ptrdesty1) = (*ptrsrcy1);
                    ptrdesty1 += 2;
                }
                ptrsrcy1 += 2;

                if (!skipDstWIndex) {
                    (*ptrdestcb1) = (*ptrsrccb1);
                    ptrdestcb1 += 4;
                }
                ptrsrccb1 += 4;

                if (!skipDstWIndex) {
                    (*ptrdestcr1) = (*ptrsrccr1);
                    ptrdestcr1 += 4;
                }
                ptrsrccr1 += 4;

                if (!skipDstWIndex) {
                    dstWIndex += 2;
                }
                srcWIndex += 2;

                calDstWIndex = (int)((srcWIndex) * dstWidth / srcWidth);
                if (calDstWIndex < dstWIndex) { // skip next dst index
                    skipDstWIndex = true;
                } else {
                    skipDstWIndex = false;
                }
            }
        } else {
            ptrsrcy1  += srcystride;
            ptrsrccb1 += srcystride;
            ptrsrccr1 += srcystride;
        }

        srcHIndex++;
        if (!skipDstHIndex) {
            dstHIndex++;
        }
        calDstHIndex = (int)((srcHIndex) * dstHeight / srcHeight);
        if (calDstHIndex < dstHIndex) { // skip next dst index
            skipDstHIndex = true;
        } else {
            skipDstHIndex = false;
        }
    }
}

void yuyv422_to_yuv420p(unsigned char* bufsrc, unsigned char* bufdest, int width, int height)
{
    unsigned char* ptrsrcy1, *ptrsrcy2;
    unsigned char* ptrsrcy3, *ptrsrcy4;
    unsigned char* ptrsrccb1, *ptrsrccb2;
    unsigned char* ptrsrccb3, *ptrsrccb4;
    unsigned char* ptrsrccr1, *ptrsrccr2;
    unsigned char* ptrsrccr3, *ptrsrccr4;
    int srcystride, srcccstride;

    ptrsrcy1  = bufsrc ;
    ptrsrcy2  = bufsrc + (width << 1) ;
    ptrsrcy3  = bufsrc + (width << 1) * 2 ;
    ptrsrcy4  = bufsrc + (width << 1) * 3 ;

    ptrsrccb1 = bufsrc + 1;
    ptrsrccb2 = bufsrc + (width << 1) + 1;
    ptrsrccb3 = bufsrc + (width << 1) * 2 + 1;
    ptrsrccb4 = bufsrc + (width << 1) * 3 + 1;

    ptrsrccr1 = bufsrc + 3;
    ptrsrccr2 = bufsrc + (width << 1) + 3;
    ptrsrccr3 = bufsrc + (width << 1) * 2 + 3;
    ptrsrccr4 = bufsrc + (width << 1) * 3 + 3;

    srcystride  = (width << 1) * 3;
    srcccstride = (width << 1) * 3;

    unsigned char* ptrdesty1, *ptrdesty2;
    unsigned char* ptrdesty3, *ptrdesty4;
    unsigned char* ptrdestcb1, *ptrdestcb2;
    unsigned char* ptrdestcr1, *ptrdestcr2;
    int destystride, destccstride;

    ptrdesty1 = bufdest;
    ptrdesty2 = bufdest + width;
    ptrdesty3 = bufdest + width * 2;
    ptrdesty4 = bufdest + width * 3;

    ptrdestcb1 = bufdest + width * height;
    ptrdestcb2 = bufdest + width * height + (width >> 1);

    ptrdestcr1 = bufdest + width * height + width * (height >> 2);
    ptrdestcr2 = bufdest + width * height + width * (height >> 2) + (width >> 1);

    destystride  = (width) * 3;
    destccstride = width >> 1;

    int i, j;

    for(j = 0; j < (height / 4); j++) {
        for(i = 0; i < (width / 2); i++) {
            (*ptrdesty1++) = (*ptrsrcy1);
            (*ptrdesty2++) = (*ptrsrcy2);
            (*ptrdesty3++) = (*ptrsrcy3);
            (*ptrdesty4++) = (*ptrsrcy4);

            ptrsrcy1 += 2;
            ptrsrcy2 += 2;
            ptrsrcy3 += 2;
            ptrsrcy4 += 2;

            (*ptrdesty1++) = (*ptrsrcy1);
            (*ptrdesty2++) = (*ptrsrcy2);
            (*ptrdesty3++) = (*ptrsrcy3);
            (*ptrdesty4++) = (*ptrsrcy4);

            ptrsrcy1 += 2;
            ptrsrcy2 += 2;
            ptrsrcy3 += 2;
            ptrsrcy4 += 2;

            (*ptrdestcr1) = (*ptrsrccr1);
            (*ptrdestcr2) = (*ptrsrccr3);
            ptrdestcr1 += 1;
            ptrdestcr2 += 1;

            ptrsrccr1 += 4;
            ptrsrccr3 += 4;

            (*ptrdestcb1) = (*ptrsrccb1);
            (*ptrdestcb2) = (*ptrsrccb3);
            ptrdestcb1 += 1;
            ptrdestcb2 += 1;

            ptrsrccb1 += 4;
            ptrsrccb3 += 4;

        }

        /* Update src pointers */
        ptrsrcy1  += srcystride;
        ptrsrcy2  += srcystride;
        ptrsrcy3  += srcystride;
        ptrsrcy4  += srcystride;

        ptrsrccb1 += srcccstride;
        ptrsrccb3 += srcccstride;

        ptrsrccr1 += srcccstride;
        ptrsrccr3 += srcccstride;

        /* Update dest pointers */
        ptrdesty1 += destystride;
        ptrdesty2 += destystride;
        ptrdesty3 += destystride;
        ptrdesty4 += destystride;

        ptrdestcb1 += destccstride;
        ptrdestcb2 += destccstride;

        ptrdestcr1 += destccstride;
        ptrdestcr2 += destccstride;
    }
}

void yuv422_to_rgb565(unsigned char* buf, unsigned char* rgb, int width, int height)
{
    static int times = 0;

    int x, y, z = 0;
    int blocks;

    blocks = (width * height) * 2;

    for (y = 0; y < blocks; y += 4) {
        unsigned char Y1, Y2, U, V;

        Y1 = buf[y + 0];
        U = buf[y + 1];
        Y2 = buf[y + 2];
        V = buf[y + 3];

        yuv422_to_rgb565(Y1, U, V, &rgb[y]);
        yuv422_to_rgb565(Y2, U, V, &rgb[y + 2]);
    }
    times++;
}

void yuv422_to_rgb565(unsigned char y, unsigned char u, unsigned char v, unsigned char* rgb)
{
    int r, g, b;
    int rgb16;

    r = (1192 * (y - 16) + 1634 * (v - 128) ) >> 10;
    g = (1192 * (y - 16) - 833 * (v - 128) - 400 * (u - 128) ) >> 10;
    b = (1192 * (y - 16) + 2066 * (u - 128) ) >> 10;

    r = r > 255 ? 255 : r < 0 ? 0 : r;
    g = g > 255 ? 255 : g < 0 ? 0 : g;
    b = b > 255 ? 255 : b < 0 ? 0 : b;

    rgb16 = (int)(((r >> 3) << 11) | ((g >> 2) << 5) | ((b >> 3) << 0));

    *rgb = (unsigned char)(rgb16 & 0xFF);
    rgb++;
    *rgb = (unsigned char)((rgb16 & 0xFF00) >> 8);
}

void yuv422_to_rgb565_32Align(unsigned char* buf, unsigned char* rgb, int width, int height)
{
    int destWidth = ((width + 31) >> 5) << 5;
    int srcblocks = width * height * 2;
    int destblocks = destWidth * height * 2;
    int pitch = 32 - width % 32;

    memset(rgb, 0, destblocks);
    int src = 0, dest = 0;
    for (; src < srcblocks && dest < destblocks; src += 4, dest += 4) {
        unsigned char Y1, Y2, U, V;

        Y1 = buf[src + 0];
        U = buf[src + 1];
        Y2 = buf[src + 2];
        V = buf[src + 3];

        if ((srcblocks != destblocks) && (dest > 0) && (dest % (destWidth * 2) % (width * 2) == 0)) {
            dest += pitch * 2;
        }

        yuv422_to_rgb565(Y1, U, V, &rgb[dest]);
        yuv422_to_rgb565(Y2, U, V, &rgb[dest + 2]);
    }
}
/*
void yuv422_to_rgb565_32Align_S(unsigned char *buf, unsigned char *Y, unsigned char *U, unsigned char *V,
    unsigned char *rgb, int width, int height)
{
    if (!(Y&&U&&V))
        return;
    int actualWidth = ((width + 31)>>5)<<5;
    yuv422decode_32Align(buf, Y, U, V, width, height);
    yuv422_to_rgb565_S(rgb, Y, U, V, actualWidth, height, yuv_to_rgb565_table);
}

void yuv422decode( unsigned char *buf,  unsigned char *Y, unsigned char*U, unsigned char* V, int width, int height)
{
    int blocks;
    int y;
    blocks = (width * height) * 2;
    for ( y = 0; y < blocks; y+=4)
    {
        int y1=0, y2=0, u=0, v=0;

        y1 = buf[y + 0];
        u = buf[y + 1];
        y2 = buf[y + 2];
        v = buf[y + 3];

        *Y++=y1;
        *U++=u;
        *Y++=y2;
        *V++=v;
    }
}

void yuv422decode_32Align( unsigned char *buf,  unsigned char *Y, unsigned char*U, unsigned char* V, int width, int height)
{
    int destWidth = ((width + 31)>>5)<<5;
    int srcblocks = width * height * 2;
    int destblocks = destWidth * height * 2;
    int pitch = 32 - width%32;

    int src=0, dest=0;
    for (; src<srcblocks && dest<destblocks; src+=4, dest+=4){
        unsigned char y1, y2, u, v;
        if ((srcblocks != destblocks) && (dest > 0) && (dest%(destWidth*2)%(width*2) == 0)){
            dest += pitch*2;
            Y+=pitch;
            U += pitch>>1;
            V += pitch>>1;
        }

        y1 = buf[src + 0];
        u = buf[src + 1];
        y2 = buf[src + 2];
        v = buf[src + 3];

         *Y++=y1;
        *U++=u;
        *Y++=y2;
        *V++=v;
    }
}
*/
