//#define LOG_NDEBUG 0
#define LOG_TAG     "Camera3VoWritebackHelper"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <unistd.h>
#include <log/log.h>
#include <utils/backtrace.h>
#include <media/stagefright/MediaCodec.h> // bad workaroind , to make Debug.h work... orz..
#include "Camera3VoWritebackHelper.h"
#include "Debug.h"

#define DUMP_VOWB_OUTPUT_BUF          ( 0)


#define OUTPUT_FRAME_COUNT_INIT_VAL   (-2)
#define DEQUEUE_MIN_SIZE              ( 2)
#define MAX_DEQUEUE_ACC_COUNT         ( 2)

static const nsecs_t kWaitNs = 5*1E9;

/**
 * - how to use pLock of input buffers?
 * - how to pass and get PTS to firmeware
 */
Camera3VoWritebackHelper::Camera3VoWritebackHelper(VOWBCallback *pCallback)
    :mInstance(NULL),
     mEnqueueThread(NULL),
     mDequeueThread(NULL),
     mInputBufsInfo(NULL),
     mOutputDumpCount(0),
     mCallback(pCallback)
{
    // do something..
    ALOGD("Camera3VoWritebackHelper()");

    for(int i=0;i<VO_WRITEBAC_BUF_COUNT;i++){
        memset(&(mBufs[i]),0x0,sizeof(vowb_allocator_info_t));
    }
}

Camera3VoWritebackHelper::~Camera3VoWritebackHelper()
{
    // do something..
    ALOGD("~Camera3VoWritebackHelper()");
    // call uninit again, for safty
    uninit();
}

status_t Camera3VoWritebackHelper::init(uint32_t bufCount, int32_t width, int32_t height)
{
    ALOGD("Init VOWriteBack start bufCount %d %dx%d",bufCount,width,height);

    CHECK(mInstance==NULL);

    mBufCount = bufCount;
    mWidth = width;
    mHeight = height;

    mAvailCounter = OUTPUT_FRAME_COUNT_INIT_VAL;
    mVoAvailCounter = VO_WRITEBAC_BUF_COUNT;

    mInstance = new VOWriteBack(bufCount);

    CHECK(mInstance->lib_rtk_codec_vo_config());
    CHECK(mInstance->lib_rtk_codec_vo_initICQRingBuffer());
    CHECK(mInstance->lib_rtk_codec_vo_initWriteBackRingBuffer());

    int32_t dbit = 0x0;
    CHECK(mInstance->lib_rtk_codec_vo_ConfigWriteBackBuffers(mWidth,mHeight,1,dbit));

    ALOGD("Init VOWriteBack complete");

    return OK;
}

status_t Camera3VoWritebackHelper::uninit()
{
    ALOGD("Uninit start");

    // clear buffers first?

    if(mInstance != NULL) {
        delete mInstance;
        mInstance = NULL;
    }

    if(mInputBufsInfo != NULL){
        ALOGD("free mInputBufsInfo %p",mInputBufsInfo);
        free(mInputBufsInfo);
        mInputBufsInfo = NULL;
    }

    ALOGD("Uninit complete");
    return OK;
}

status_t Camera3VoWritebackHelper::setBufs(allocator_info_t *pBufs, int32_t size)
{
    CHECK(size==VO_WRITEBAC_BUF_COUNT);

    // first prepare mBufs
    for(int i=0;i<size;i++){
        mBufs[i].alloc_info = &(pBufs[i]);
        // default status is cleared
        mBufs[i].status = STATUS_CLEARED;
        ALOGV("[%d] allocator %p",i,mBufs[i].alloc_info->allocator);
    }

    // second pass info to VO
    for(int i=0;i<size;i++) {
        uint32_t phys = (uint32_t) mBufs[i].alloc_info->phyAddr;
        int32_t index = i;
        mInstance->lib_rtk_codec_vo_AddWriteBackBuffers(index, (unsigned long)phys, 0, 0);
        ALOGV("[%s] Allocate write back buffers %d with physical address %p done",
            __FUNCTION__,
            index,
            (unsigned long)phys);

#if 0
        mInstance->lib_rtk_codec_vo_LockWriteBackBuffers((unsigned long)phys, 0);
        ALOGD("Lock write back buffer %p [%s]",
            (unsigned long)phys,
            (ret?"Success":"failed"));
#endif
    }

    return OK;
}

status_t Camera3VoWritebackHelper::setInputBufsInfo(allocator_info_t *pBufs, int32_t size)
{
    CHECK(mInputBufsInfo==NULL);
    mInputBufsInfo = (allocator_info_t **) malloc(size *sizeof(allocator_info_t *));
    for(int i=0;i<size;i++){
        mInputBufsInfo[i] = &(pBufs[i]);

        uint8_t *pLockVAddr = NULL;
        // fill up VOWB extension fields
        mInputBufsInfo[i]->pLockAddr =
            mInstance->lib_rtk_codec_vo_getpLockAddrWithIdx(i, &pLockVAddr);
        CHECK(pLockVAddr != NULL);
        mInputBufsInfo[i]->yAddr = (int64_t) mInputBufsInfo[i]->phyAddr;
        mInputBufsInfo[i]->uAddr = mInputBufsInfo[i]->yAddr +
            (mInputBufsInfo[i]->width_aligned * mInputBufsInfo[i]->height_aligned);

        mInputBufsInfo[i]->pLockVAddr = (int8_t *) pLockVAddr;
        mInputBufsInfo[i]->pRecvVAddr = mInputBufsInfo[i]->pLockVAddr + 1;

        ALOGV("InputBufInfo [%d] %dx%d pLock %.8x yAddr %.8x uAddr %.8x",
            i,
            mInputBufsInfo[i]->width,
            mInputBufsInfo[i]->height,
            mInputBufsInfo[i]->pLockAddr,
            mInputBufsInfo[i]->phyAddr, //yAddr,
            mInputBufsInfo[i]->uAddr);
    }

    return OK;
}

/**
 * enqueue v4l2 buffer to VO to do deinterlace
 * called in mEnqueueThread
 */
status_t Camera3VoWritebackHelper::enqueueInputBuf2VO()
{

    CHECK(mEnqueuedBuffers.size()>=2);

    // retrive 2 input buffers from queue
    allocator_info_t *pBuf0 = mEnqueuedBuffers[0];
    allocator_info_t *pBuf1 = mEnqueuedBuffers[1];

    // only remove the first one
    {
        Mutex::Autolock l(mLockEQ);
        mEnqueuedBuffers.removeItemsAt(0);
    }

    CHECK(pBuf0 != NULL);
    CHECK(pBuf1 != NULL);

    int width = pBuf0->width;
    int height = pBuf0->height;

    bool ret;

#if 0
    *pBuf0->pRecvVAddr = 0;
    *pBuf0->pLockVAddr = 0;
    *pBuf1->pRecvVAddr = 0;
    *pBuf1->pLockVAddr = 0;
#endif

    ALOGV("[VOWBD] 1 R:%d L:%d ; 2 R:%d L:%d",
        *pBuf0->pRecvVAddr,
        *pBuf0->pLockVAddr,
        *pBuf1->pRecvVAddr,
        *pBuf1->pLockVAddr);

    ret = mInstance->lib_rtk_codec_vo_prepareBuffer_deinterlace(
            (uint32_t) pBuf0->yAddr,     //prev.yAddr,
            (uint32_t) pBuf0->uAddr,     //prev.uAddr,
            (uint32_t) pBuf1->yAddr,     //(unsigned int)yAddr,
            (uint32_t) pBuf1->uAddr,     //(unsigned int)uAddr,
            width,
            height,
            pBuf0->pLockAddr,            //prev.pLockAddr,
            pBuf1->pLockAddr,            //pLockAddr,
            0);

    ret = mInstance->lib_rtk_codec_vo_prepareBuffer_deinterlace(
            (uint32_t) pBuf0->yAddr,     //prev.yAddr,
            (uint32_t) pBuf0->uAddr,     //prev.uAddr,
            (uint32_t) pBuf1->yAddr,     //(unsigned int)yAddr,
            (uint32_t) pBuf1->uAddr,     //(unsigned int)uAddr,
            width,
            height,
            pBuf0->pLockAddr,            //prev.pLockAddr,
            pBuf1->pLockAddr,            //pLockAddr,
            1);

    // then send Inband command twice
    int32_t cmd = 39;
    ret = mInstance->lib_rtk_process_InbandCmd(cmd);
    ALOGV("[%s] Process_InbandCmd %d", (ret == 1? "Success":"Fail"), cmd);
    ret = mInstance->lib_rtk_process_InbandCmd(cmd);
    ALOGV("[%s] Process_InbandCmd %d", (ret == 1? "Success":"Fail"), cmd);

    // assume that VOWB can produce 2 frames everytime a new frame is enqueued
    addCounter(2);
    // notify dequeue thread can move on..
    mDequeueThread->signal();

    // pass pBUf0 to return candidate queue
    enqueueRetBuf(pBuf0);

    return OK;
}


/**
 * check if there is any buffer is used completed by VO,
 * and return it back to V4L2
 * called in mDequeueThread
 */
status_t Camera3VoWritebackHelper::returnBufToV4L2()
{
#if 1
    int loop = 0;

    //dumpAcquiredInputBufferNums();
    dumpRetBufQueueState();

    while (mReturnBuffers.size() > 0 && loop < 10) {

        allocator_info_t *pBuf = peekRetBuf();

        CHECK(pBuf != NULL);

        int8_t received = *(pBuf->pRecvVAddr);
        int8_t locked = *(pBuf->pLockVAddr);

        if(received==1 && locked==0) {
            ALOGV("Return buf %d to V4L2",pBuf->index);
            mCallback->returnSrcBuffer(pBuf);
            dequeueRetBuf();
            break;
        }else{
            ALOGV("[XXXX] buffer %d is locked [R:%d L:%d]",
                pBuf->index,
                received,
                locked);
            //usleep(1E3);
            loop++;
        }
    }
    return OK;
#else
    while(mReturnBuffers.size() > 2) {
        ALOGD("[XXXX] returnBufToV4L2 %d",mReturnBuffers.size());
        allocator_info_t *pBuf = peekRetBuf();
        mCallback->returnSrcBuffer(pBuf);
        dequeueRetBuf();
    }
    return OK;
#endif
}

status_t Camera3VoWritebackHelper::enqueueInputBuf(allocator_info_t *pBuf)
{
    {
        Mutex::Autolock l(mLockEQ);
        mEnqueuedBuffers.push_back(pBuf);
    }

    ALOGV("[PTSVOWB] > enqueueInputBuf pts %.3f [%d %d] avail %d",
        (float)pBuf->timeStamp/(float)1E9,
        mEnqueuedBuffers.size(),
        mDequeuedBuffers.size(),
        mAvailCounter);

    enqueuePTS(pBuf->timeStamp);

    // notify thread if it is waiting..
    if(mEnqueueThread.get() != NULL) {
        mEnqueueThread->signal();
    }

    if(mDequeueThread.get() != NULL) {
        mDequeueThread->signal();
    }

    return OK;
}

status_t Camera3VoWritebackHelper::enqueuePTS(nsecs_t pts)
{
    Mutex::Autolock l(mLockPTSQ);
    mPTSQueue.push_back(pts);
    return OK;
}

status_t Camera3VoWritebackHelper::dequeuePTS()
{
    Mutex::Autolock l(mLockPTSQ);
    mPTSQueue.removeItemsAt(0);
    return OK;
}

status_t Camera3VoWritebackHelper::enqueueDequeuedBuffers(allocator_info_t *pBuf)
{
    Mutex::Autolock l(mLockDQ);
    mDequeuedBuffers.push_back(pBuf);
    return OK;
}

allocator_info_t * Camera3VoWritebackHelper::dequeueOutputBuf(bool blocked)
{
    ALOGV("dequeueOutputBuf start");
    if(blocked) {
        while(mDequeuedBuffers.size()<=0){
            usleep(1E3);
        }
    }else{
        if(mDequeuedBuffers.size()<=0){
            return NULL;
        }
    }

    allocator_info_t *rst = mDequeuedBuffers[0];

    ALOGV("dequeueOutputBuf cont");
    {
        Mutex::Autolock l(mLockDQ);
        mDequeuedBuffers.removeItemsAt(0);
    }

    ALOGV("dequeueOutputBuf done");
    // and clear this buffer, so VOWB can use it again
    mDequeueThread->signal();

    ALOGV("[PTSVOWB]      < %.3f [DQ:%d]",
        (float)rst->timeStamp/(float)1E9,
        mDequeuedBuffers.size());

    return rst;
}

status_t Camera3VoWritebackHelper::releaseOutputBuf(allocator_info_t *pBuf)
{
    CHECK(pBuf != NULL);
    bool ret;

    ret = mInstance->lib_rtk_codec_vo_ClearWriteBackBuffers(
            (unsigned long)pBuf->phyAddr,
            0);

    ALOGV("[PTSVOWB] releaseOutputBuf idx %d ret %d dequeueBufSize [%d]",
        pBuf->index,
        ret,
        mDequeuedBuffers.size());

    addVoCounter(1);

    if(mVoAvailCounter>=DEQUEUE_MIN_SIZE){
        ALOGV("[%s] Notify dequeueThread",__FUNCTION__);
        mDequeueThread->signal();
    }

    return OK;
}

/**
 * start inner worker threads
 */
status_t Camera3VoWritebackHelper::start()
{
    // start dequeue thread first
    CHECK(mDequeueThread.get()==NULL);
    mDequeueThread = new VODequeueThread(this);
    mDequeueThread->run("DequeueVO");

    CHECK(mEnqueueThread.get()==NULL);
    mEnqueueThread = new VOEnqueueThread(this);
    mEnqueueThread->run("EnqueueVO");

    return OK;
}

status_t Camera3VoWritebackHelper::stop()
{
    ALOGV("stop() start");

    if(mEnqueueThread.get() != NULL) {
        mEnqueueThread->stop();
        mEnqueueThread->signal();
        mEnqueueThread->requestExitAndWait();
        mEnqueueThread.clear();
        mEnqueueThread = NULL;
    }

    ALOGV("stop() cont");

    if(mDequeueThread.get() != NULL) {
        mDequeueThread->stop();
        mDequeueThread->signal();
        mDequeueThread->requestExitAndWait();
        mDequeueThread.clear();
        mDequeueThread = NULL;
    }

    ALOGV("stop() done");
    return OK;
}

status_t Camera3VoWritebackHelper::dequeueOutputBufFromVO(YUV_STATE *pBuf0, YUV_STATE *pBuf1)
{
    CHECK(mAvailCounter>=DEQUEUE_MIN_SIZE);

    ALOGV("dequeueOutputBufFromVO mAvailCounter %d",mAvailCounter);

    memset(pBuf0,0x0,sizeof(YUV_STATE));
    memset(pBuf1,0x0,sizeof(YUV_STATE));
    // get 2 frames from VOWB
    // use blocking mode
    int loop = 10;
    bool rst=false;
    for(int i=0;i<loop;i++){
        rst = mInstance->lib_rtk_codec_vo_getWriteBackFrame(pBuf0);
        if(rst){
            break;
        }else{
            usleep(30);
            ALOGW("retry [%d] loop %d",__LINE__,i);
        }
    }
    CHECK(rst);

    rst = false;
    for(int i=0;i<loop;i++){
        rst = mInstance->lib_rtk_codec_vo_getWriteBackFrame(pBuf1);
        if(rst){
            break;
        }else{
            usleep(30);
            ALOGW("retry [%d] loop %d",__LINE__,i);
        }

    }
    CHECK(rst);

    // then decrease mAvailCounter
    addCounter(-2);
    addVoCounter(-2);

    return OK;
}

void Camera3VoWritebackHelper::addVoCounter(int val)
{
    Mutex::Autolock l(mVoCounterLock);
    mVoAvailCounter = mVoAvailCounter + val;
}

void Camera3VoWritebackHelper::addCounter(int val)
{
    Mutex::Autolock l(mCounterLock);
    mAvailCounter = mAvailCounter + val;
}

allocator_info_t * Camera3VoWritebackHelper::findVOBufByYuvState(YUV_STATE *pBuf)
{
    for(int i=0;i<VO_WRITEBAC_BUF_COUNT;i++) {
        uint32_t phyAddr = (uint32_t) mBufs[i].alloc_info->phyAddr;
        if(phyAddr == pBuf->Y_addr){
            return mBufs[i].alloc_info;
        }
    }
    return NULL;
}

status_t Camera3VoWritebackHelper::enqueueRetBuf(allocator_info_t *pBuf)
{
    Mutex::Autolock l(mLockRetQ);
    mReturnBuffers.push_back(pBuf);
    return OK;
}

status_t Camera3VoWritebackHelper::dequeueRetBuf()
{
    Mutex::Autolock l(mLockRetQ);
    mReturnBuffers.removeItemsAt(0);
    return OK;
}


allocator_info_t * Camera3VoWritebackHelper::peekRetBuf()
{
    Mutex::Autolock l(mLockRetQ);
    CHECK(mReturnBuffers.size()>0);
    allocator_info_t *rst;
    rst = mReturnBuffers[0];
    return rst;
}

// ---------------------------- VOEnqueueThread ------------------------------- //
Camera3VoWritebackHelper::VOEnqueueThread::VOEnqueueThread(Camera3VoWritebackHelper *parent)
    :mParent(parent),
     mRunning(true)
{
    // do something
}

Camera3VoWritebackHelper::VOEnqueueThread::~VOEnqueueThread()
{
    // do something
    ALOGD("~VOEnqueueThread()");
}

bool Camera3VoWritebackHelper::VOEnqueueThread::threadLoop()
{
    if(!mRunning) {
        return false;
    }

    int minEnqueuedBufSize = 2;

    // wait input buffers ready
    int loop = 0;
    while(mParent->mEnqueuedBuffers.size()<minEnqueuedBufSize && mRunning && loop < 10){
        // wait buffer ready until terminated or nums of buffer is enough
        Mutex::Autolock l(mParent->mEnqueueLock);
        mParent->mEnqueueSignal.waitRelative(mParent->mEnqueueLock,kWaitNs);
        loop++;

        if(loop >= 10 && mParent->mEnqueuedBuffers.size()<minEnqueuedBufSize){
            ALOGW("VOEnqueueThread No input buffer %d",mParent->mEnqueuedBuffers.size());
            return true;
        }
    }

    if(mParent->mEnqueuedBuffers.size() >= minEnqueuedBufSize) {
        mParent->enqueueInputBuf2VO();
    }

    return true;
}

void Camera3VoWritebackHelper::VOEnqueueThread::signal()
{
    Mutex::Autolock l(mParent->mEnqueueLock);
    mParent->mEnqueueSignal.signal();
}

// ---------------------------- VODequeueThread ------------------------------- //
Camera3VoWritebackHelper::VODequeueThread::VODequeueThread(Camera3VoWritebackHelper *parent)
    :mParent(parent),
     mRunning(true)
{
    // do something
}

Camera3VoWritebackHelper::VODequeueThread::~VODequeueThread()
{
    // do something
    ALOGD("~VODequeueThread()");
}

bool Camera3VoWritebackHelper::VODequeueThread::threadLoop()
{
    ALOGV("PTSVOWB - VODequeueThread launch - P [%d %d]",mParent->mAvailCounter,mParent->mDequeuedBuffers.size());

    if(!mRunning) {
        return false;
    }

    int accumulatedDequeueBufsMax = MAX_DEQUEUE_ACC_COUNT;
    int loop = 0;
    // wait
    while(mRunning &&
        (mParent->mAvailCounter < DEQUEUE_MIN_SIZE ||
         mParent->mVoAvailCounter < DEQUEUE_MIN_SIZE ||
         mParent->mDequeuedBuffers.size() > accumulatedDequeueBufsMax) ) {

        ALOGV("VODequeueThread wait start [%d]",mParent->mDequeuedBuffers.size());
        Mutex::Autolock l(mParent->mDequeueLock);
        mParent->mDequeueSignal.waitRelative(mParent->mDequeueLock,kWaitNs);
        ALOGV("[PTSVO]         VODequeueThread launch - P W [DQ:%d,AV:%d]",
            mParent->mDequeuedBuffers.size(),
            mParent->mAvailCounter);
        loop++;

#if 1
        if(loop >= 5 &&
            (mParent->mAvailCounter < DEQUEUE_MIN_SIZE ||
             mParent->mVoAvailCounter < DEQUEUE_MIN_SIZE ||
             mParent->mDequeuedBuffers.size() > accumulatedDequeueBufsMax) ) {

            ALOGV("[VOWBD] VODequeueThread No output buffer [%d][%d,%d]",
                mParent->mDequeuedBuffers.size(),
                mParent->mAvailCounter,
                mParent->mVoAvailCounter);
            return true;
        }
#endif
    }

    if(mParent->mAvailCounter >= DEQUEUE_MIN_SIZE && mParent->mVoAvailCounter >= DEQUEUE_MIN_SIZE) {

        ALOGV("PTSVOWB - VODequeueThread launch - E");

        YUV_STATE buf0;
        YUV_STATE buf1;

        mParent->dequeueOutputBufFromVO(&buf0,&buf1);

        allocator_info_t *pInfo0 = mParent->findVOBufByYuvState(&buf0);
        allocator_info_t *pInfo1 = mParent->findVOBufByYuvState(&buf1);

        CHECK(pInfo0 != NULL);
        CHECK(pInfo1 != NULL);

        // then calculate result pts
        ALOGV("Current PTS Queue size %d",mParent->mPTSQueue.size());

        CHECK(mParent->mPTSQueue.size()>=2);

        nsecs_t pts0 = mParent->mPTSQueue[0];
        nsecs_t pts1 = (mParent->mPTSQueue[1] + mParent->mPTSQueue[0])/2;

        mParent->dequeuePTS();

        // set pts
        pInfo0->timeStamp = pts0;
        pInfo1->timeStamp = pts1;

#if (DUMP_VOWB_OUTPUT_BUF)
        if(mParent->mOutputDumpCount < 10) {
            char path[64] = {'\0'};
            sprintf(path,"/tmp/VOWB-%d.raw",mParent->mOutputDumpCount);
            mParent->dumpData(pInfo0->vAddr, pInfo0->size, path, &mParent->mOutputDumpCount);
            sprintf(path,"/tmp/VOWB-%d.raw",mParent->mOutputDumpCount);
            mParent->dumpData(pInfo1->vAddr, pInfo1->size, path, &mParent->mOutputDumpCount);
        }
#endif
        // then, add to dequeue queue
        mParent->enqueueDequeuedBuffers(pInfo0);
        mParent->enqueueDequeuedBuffers(pInfo1);

        // notify V4L2 that buffer is available
        mParent->mCallback->signalBufferAvailable();

        /* cause some new output frames are came out, so check if there is
         * any available input buffer can return back to v4l2
         */
        mParent->returnBufToV4L2();

        ALOGV("PTSVOWB - VODequeueThread launch - X");

    }

    return true;
}

void Camera3VoWritebackHelper::VODequeueThread::signal()
{
    Mutex::Autolock l(mParent->mDequeueLock);
    mParent->mDequeueSignal.signal();
}

// ------------------------- Debug helper function ------------------------------ //

void Camera3VoWritebackHelper::dumpData(void *vaddr, int32_t size, char *path, uint32_t *pCnt)
{
    FILE *fd = fopen(path,"w");
    if(fd != NULL) {
        fwrite(vaddr,1,size,fd);
        int rst = fclose(fd);

        ALOGV("[%s] path %s size %d rst %d",
            __FUNCTION__,
            path,
            size,
            rst);

        (*pCnt)++;
    }
}

void Camera3VoWritebackHelper::dumpAcquiredInputBufferNums()
{
    int32_t inEnqueueQueue = mEnqueuedBuffers.size();
    int32_t inRetQueue = mReturnBuffers.size();

    ALOGV("Buffer acquired by VOWB %d %d",inEnqueueQueue, inRetQueue);
}

void Camera3VoWritebackHelper::dumpRetBufQueueState()
{
    int32_t size = mReturnBuffers.size();
    for(int i=0;i<size;i++) {
        allocator_info_t *pBuf = mReturnBuffers[i];
        int8_t received = *(pBuf->pRecvVAddr);
        int8_t locked = *(pBuf->pLockVAddr);
        ALOGV("[RETV4L2] ID:%d R:%d L:%d",
            pBuf->index,
            received,
            locked);
    }
}

