// #define LOG_NDEBUG 0
#define LOG_TAG "JPEGCompressHelper"

#include <iostream>
#include <fstream>
#include <vector>
#include <cstdint>
#include <memory>

#include <log/log.h>
#include <Debug.h>

#include <jpeglib.h>

using namespace std;

static inline int32_t findUOffset(int32_t w, int32_t h, int32_t x, int32_t y)
{
    int32_t baseOffset = w*h;
    baseOffset=baseOffset+(y/2)*w;
    baseOffset=baseOffset+((x/2)*2);
    return baseOffset;
}

static int jpgCount = 0;

static inline void saveFile(void *vaddr,
        int size,
        char *path)
{
    LOG_FUNCTION_NAME();
    FILE *fd = fopen(path,"w");
    if(fd != NULL) {
        fwrite(vaddr,1,size,fd);
        int rst = fclose(fd);
        ALOGV("[%s] path %s size %d rst %d",__FUNCTION__,path,size,rst);
    }
}

int32_t compressYUVtoJPEG(
    uint8_t *input,
    const int width,
    const int height,
    uint8_t *out,
    int32_t outBufferSize,
    int32_t quality) {
    LOG_FUNCTION_NAME();
    // width and height checking
    CHECK(width%2==0);
    CHECK(height%2==0);

    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    JSAMPROW row_ptr[1];
    int row_stride;

    unsigned long outlen = 0;
    uint8_t *outbuffer = NULL;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_mem_dest(&cinfo, &outbuffer, &outlen);

    cinfo.image_width = width & -1;
    cinfo.image_height = height & -1;
    cinfo.input_components = 3;
    cinfo.in_color_space = JCS_YCbCr; //libJPEG expects YUV 3bytes, 24bit

    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, quality, TRUE);
    jpeg_start_compress(&cinfo, TRUE);

    // allocate temp buffer
    vector<uint8_t> tmprowbuf(width * 3);

    JSAMPROW row_pointer[1];
    row_pointer[0] = &tmprowbuf[0];

    // loop each row
    while(cinfo.next_scanline < cinfo.image_height) {

        // i: input buffer index (horizontal)
        // j: meta buffer index
        uint32_t i, j;
        uint32_t offsetY = cinfo.next_scanline * cinfo.image_width;

        for(i=0, j=0 ; i < cinfo.image_width ; i+=1, j+=3) {
            // Y
            tmprowbuf[j+0] = input[offsetY+i];

            int x = i;
            int y = cinfo.next_scanline;
            int32_t uOffset = findUOffset(
                                width,
                                height,
                                i,
                                cinfo.next_scanline);

            // U
            tmprowbuf[j+1] = *(input + uOffset);
            // V
            tmprowbuf[j+2] = *(input + uOffset + 1);
        }
        jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }

    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);

    ALOGD("%s outlen : %ld buffer %p", __FUNCTION__, outlen, outbuffer);

#if 0
    char path[128];
    memset(path,0x0,128);
    sprintf(path, "/tmp/camera3-%d.jpg",jpgCount);

    saveFile(outbuffer, outlen, path);
    jpgCount++;
#endif

    CHECK((int32_t)outlen <= outBufferSize);
    if(outbuffer != NULL) {
        memcpy(out,outbuffer,outlen);
        free(outbuffer);
    }
    return (int32_t) outlen;
}
